###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from sklearn.ensemble import RandomForestClassifier
import sklearn.metrics as metrics
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression, LinearRegression
import math
from sklearn.model_selection import train_test_split


from sciutil import SciUtil
from scivae.util import convert_str_labels_to_ints
import numpy as np

class Validate(object):

    def __init__(self, vae, class_labels: list, train_percentage=85.0, random_state=17, sciutil=None):
        self.u = sciutil if sciutil is not None else SciUtil()
        self.vae = vae
        self.class_labels = convert_str_labels_to_ints(class_labels)
        self.train_split = train_percentage/100.0
        self.data = vae.get_encoded_data()
        self.X_train, self.X_test, self.y_train, self.y_test = None, None, None, None
        # Generate the training and testing data
        self.random_state = random_state
        self.generate_test_train()

    def generate_test_train(self):
        # remove any NAs from the dataset
        self.data[np.isneginf(self.data)] = 0
        self.data[np.isinf(self.data)] = 0
        self.data[np.isnan(self.data)] = 0

        train_size = int(math.ceil(len(self.data) * self.train_split)) - 1
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.data, self.class_labels,
                                                                                random_state=self.random_state,
                                                                                train_size=train_size)

    def predict(self, method: str, metric: str, arg=None) -> float:
        classifier = None
        if method == 'svm':
            classifier = self.svm()
        elif method == 'rf':
            classifier = self.rf(arg)
        elif method == 'lin_reg':
            classifier = self.lin_reg()
            classifier.fit(self.X_train, self.y_train)
            return classifier.score(self.X_test, self.y_test)
        elif method == 'log_reg':
            classifier = self.log_reg()
            classifier.fit(self.X_train, self.y_train)
            return classifier.score(self.X_test, self.y_test)

        classifier.fit(self.X_train, self.y_train)
        predicted = classifier.predict(self.X_test)
        if metric == 'accuracy':
            return metrics.accuracy_score(self.y_test, predicted)
        elif metric == 'balanced_accuracy':
            return metrics.balanced_accuracy_score(self.y_test, predicted)
        elif metric == 'average_precision':
            return metrics.average_precision_score(self.y_test, predicted)
        elif metric == 'neg_brier_score':
            return metrics.brier_score_loss(self.y_test, predicted)
        elif metric == 'f1':
            return metrics.f1_score(self.y_test, predicted)
        elif metric == 'roc_auc':
            return metrics.roc_auc_score(self.y_test, predicted)
        else:
            msg = self.u.msg.msg_arg_err("Validate.predict", "metric", metric, ['accuracy',
                                                                                'balanced_accuracy',
                                                                                'average_precision', 'neg_brier_score',
                                                                                'f1', 'roc_auc'])
            self.u.err_p([msg])
            raise Exception(msg)

    def rf(self, n_trees=10):
        n_trees = n_trees if n_trees is not None else 10
        return RandomForestClassifier(n_estimators=n_trees, random_state=self.random_state)

    def log_reg(self):
        return LogisticRegression(random_state=self.random_state)

    def lin_reg(self):
        return LinearRegression()

    def svm(self):
        return SVC(random_state=self.random_state)



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from sciutil import SciUtil, SciException

from tensorflow.keras import backend as K
import tensorflow as tf
import tensorflow_probability as tfp


class LossException(SciException):
    def __init__(self, message=''):
        Exception.__init__(self, message)


class Loss:

    """
    References:
        https://github.com/pren1/keras-MMD-Variational-Autoencoder/blob/master/Keras_MMD_Variational_Autoencoder.ipynb
        https://github.com/s-omranpour/X-VAE-keras/blob/master/VAE/VAE_MMD.ipynb
        https://github.com/ShengjiaZhao/MMD-Variational-Autoencoder/blob/master/mmd_vae.py
        https://github.com/CancerAI-CL/IntegrativeVAEs/blob/master/code/models/mmvae.py
        https://keras.io/guides/training_with_built_in_methods/
    """

    def __init__(self, loss_type: str, distance_metric: str, mmd_weight: float, multi_loss=None, sciutil=None,
                 mmcd_method='k',  beta=1.0, input_size=None):
        """

        Parameters
        ----------
        loss_type
        distance_metric
        learning_rate
        sciutil
        """
        self.u = sciutil if sciutil is not None else SciUtil()
        self.mmd_weight = mmd_weight

        self.loss_types = ['ce', 'mse', 'cor', 'cor-mse', 'multi', 'mae']
        if loss_type not in self.loss_types:
            msg = self.u.msg.msg_arg_err("Loss __init__", "loss", loss_type, self.loss_types)
            self.u.err_p([msg])
            raise LossException(msg)
        self.loss_type = loss_type
        self.distance_types = ['kl', 'mmd', 'mmcd', 'kl-mmd', 'bmmd']
        if self.loss_type == 'multi' and multi_loss is None:
            msg = self.u.msg.msg_arg_err("Loss __init__", "multi", multi_loss, 'Multi loss requires a dict of losss '
                                                                              'and corresponding idxs: {mse: [0,3]}')
            self.u.err_p([msg])
            raise LossException(msg)
        self.multi_loss_fn = multi_loss
        if distance_metric not in self.distance_types:
            msg = self.u.msg.msg_arg_err("Loss __init__", "distance_metric", distance_metric, self.distance_types)
            self.u.err_p([msg])
            raise LossException(msg)
        if distance_metric == 'mmd':
            self.mmd_weight = self.mmd_weight * 1.0
            # Check weight parameter is correct format otherwise throw error
            if not isinstance(self.mmd_weight, float) or mmd_weight < 0:
                msg = self.u.msg.msg_arg_err("Loss __init__", "mmd_weight", mmd_weight, ['>0', 'must be a float'])
                self.u.err_p([msg])
                raise LossException(msg)
        self.distance_metric = distance_metric
        self.beta = beta or 1.0
        self.mmcd_method = mmcd_method
        self.sizes = input_size
        if isinstance(input_size, list):
            total = 0
            for s in input_size:
                total += s
            self.sizes = [s/total for s in input_size]

    def get_loss(self, inputs_x, outputs_y, latent_z, latent_z_mean, latent_z_log_sigma) -> float:
        """
        Get the loss as defined by the configuration setup.
        Returns
        -------

        """
        distance = 1.0
        # Get the distance between the predicted distribution and the expected dist
        if self.distance_metric == 'kl':
            distance = self.get_kl_distance(latent_z_mean, latent_z_log_sigma)
        elif self.distance_metric == 'mmd':
            distance = self.get_mmd_distance(latent_z)
        elif self.distance_metric == 'bmmd':
            distance = self.get_bimodal_mmd_distance(latent_z)
        elif self.distance_metric == 'mmcd':
            istance = self.get_mmcd_distance(latent_z, latent_z_mean, latent_z_log_sigma, self.mmcd_method)
        elif self.distance_metric == 'kl-mmd':
            distance = self.get_mmd_distance(latent_z) + self.get_kl_distance(latent_z_mean, latent_z_log_sigma)
        else:
            msg = self.u.msg.msg_arg_err("Loss __init__", "distance_metric", self.distance_metric, self.distance_types)
            self.u.err_p([msg])
            raise LossException(msg)
        reconstruction_loss = 1000000000
        # Get reconstruction loss
        if self.loss_type == 'ce':
            reconstruction_loss = self.get_binary_crossentropy_loss(inputs_x, outputs_y)
        elif self.loss_type == 'mse':
            reconstruction_loss = self.get_mean_squared_error_loss(inputs_x, outputs_y)
        elif self.loss_type == 'cor':
            reconstruction_loss = self.get_correlation_loss(inputs_x, outputs_y)
        elif self.loss_type == 'cor-mse':
            reconstruction_loss = self.get_correlation_mse_loss(inputs_x, outputs_y)
        elif self.loss_type == 'mae':
            reconstruction_loss = self.get_mean_absolute_error_loss(inputs_x, outputs_y)
        elif self.loss_type == 'multi':
            reconstruction_loss = 0
            # ToDo: Modularise.
            loss_idx = 0
            for loss_method in self.multi_loss_fn:
                if loss_method == 'ce':
                    reconstruction_loss += self.get_binary_crossentropy_loss(inputs_x[loss_idx],
                                                                             outputs_y[loss_idx])
                elif loss_method == 'mse':
                    # normalise to the number of features
                    reconstruction_loss += self.get_mean_squared_error_loss(inputs_x[loss_idx],
                                                                            outputs_y[loss_idx])*self.sizes[loss_idx]
                elif loss_method == 'cor':
                    reconstruction_loss += self.get_correlation_loss(inputs_x[loss_idx],
                                                                     outputs_y[loss_idx])*self.sizes[loss_idx]
                elif loss_method == 'mae':
                    reconstruction_loss += self.get_mean_absolute_error_loss(inputs_x[loss_idx],
                                                                             outputs_y[loss_idx])
                loss_idx += 1
        else:
            msg = self.u.msg.msg_arg_err("Loss __init__", "loss", self.loss_type, self.loss_types)
            self.u.err_p([msg])
            raise LossException(msg)
        # Return the mean between the reconstruction loss and the learning rate * the distance between the distributions
        return K.mean(reconstruction_loss + (self.mmd_weight * distance))

    @staticmethod
    def get_binary_crossentropy_loss(input_x, output_y):
        """
        From: https://github.com/CancerAI-CL/IntegrativeVAEs/blob/master/code/models/common.py
        Parameters
        ----------
        train_inputs
        train_reconstruction

        Returns
        -------

        """
        return K.sum(K.binary_crossentropy(input_x, output_y, from_logits=True), axis=1)

    @staticmethod
    def get_mean_squared_error_loss(input_x, output_y):
        """ https://github.com/CancerAI-CL/IntegrativeVAEs/blob/master/code/models/common.py """
        return K.sum(K.square(input_x - output_y), axis=1)

    @staticmethod
    def get_mean_absolute_error_loss(input_x, output_y):
        """ MAE """
        return K.sum(K.abs(input_x - output_y), axis=1)


    @staticmethod
    def get_correlation_loss(input_x, output_y):
        """ Check this function! WIP.
        https://stackoverflow.com/questions/46619869/how-to-specify-the-correlation-coefficient-as-the-loss-function-in-keras/46620771 """
        mx = K.mean(input_x)
        my = K.mean(output_y)
        xm, ym = input_x - mx, output_y - my
        r_num = K.sum(tf.multiply(xm, ym))
        r_den = K.sqrt(tf.multiply(K.sum(K.square(xm)), K.sum(K.square(ym))))
        r = r_num / r_den

        r = K.maximum(K.minimum(r, 1.0), -1.0)
        return 1 - r

    def get_correlation_mse_loss(self, input_x, output_y):
        """ WIP. """
        return 0.5 * (self.get_correlation_loss(input_x, output_y) + self.get_mean_squared_error_loss(input_x, output_y))

    @staticmethod
    def get_correlation_loss_per_dataype(input_x, output_y):
        """ WIP. """
        mx = K.mean(input_x)
        my = K.mean(output_y)
        xm, ym = input_x - mx, output_y - my
        r_num = K.sum(tf.multiply(xm, ym))
        r_den = K.sqrt(tf.multiply(K.sum(K.square(xm)), K.sum(K.square(ym))))
        r = r_num / r_den

        r = K.maximum(K.minimum(r, 1.0), -1.0)
        return 1 - r

    @staticmethod
    def get_kl_distance(z_mean, z_log_sigma):
        """ Resources:
        https://keras.io/examples/generative/vae/
        https://github.com/geyang/variational_autoencoder_pytorch
        """
        # KL regularizer. this is the KL of q(z|x) given that the target distribution is N(0,1)
        kl_loss = 1 + z_log_sigma - K.square(z_mean) - K.exp(z_log_sigma)
        kl_loss = K.sum(kl_loss, axis=-1)
        kl_loss *= -0.5

        return kl_loss

    def get_mmcd_distance(self, latent_z, latent_z_mean, latent_z_log_sigma, method='k'):
        """
        WIP.
        https://wis.kuleuven.be/statdatascience/robust/software
        https://www.researchgate.net/publication/343471031_Outlier_detection_in_non-elliptical_data_by_kernel_MRCD
        https://wis.kuleuven.be/statdatascience/robust/papers/publications-2018/hubertdebruynerousseeuw-mcdext-wires-2018.pdf

        This code was translated from the paper: 10.1007/s11063-019-10090-0
        Maximum Mean and Covariance Discrepancy for Unsupervised Domain Adaptation
        by Wenju Zhang, Xiang Zhang, Long Lan & Zhigang Luo

        Code was in Matlab provided by Wenju Zhang accessed from github on 07/11/2020
        https://github.com/wj-zhang/McDA

        https://stats.stackexchange.com/questions/14673/measures-of-similarity-or-distance-between-two-covariance-matrices
        SchreursEtAl_KMRCD_arXiv_v1_2020.pdf

        """
        beta = self.beta or 1.0

        batch_size = K.shape(latent_z)[0]
        latent_dim = K.int_shape(latent_z)[1]
        loss_mmcd = 0  # If they don't set a proper param, then we just run with no MMCD i.e returning 0.

        # Randomly sample from a normal distribution
        true_samples = K.random_normal(shape=(batch_size, latent_dim), mean=0., stddev=1.)

        if method == 'k':  # If we're running mmcd on a kernal
            # Compute loss between the training latent space and normal data
            x_kernel = self.compute_kernel(true_samples, true_samples)
            y_kernel = self.compute_kernel(latent_z, latent_z)
            xy_kernel = self.compute_kernel(true_samples, latent_z)
            loss_mmd = K.mean(x_kernel) + K.mean(y_kernel) - 2 * K.mean(xy_kernel)
            cov_dif = K.sqrt(K.sum(K.square(tfp.stats.covariance(x_kernel) - tfp.stats.covariance(y_kernel))))
            loss_mmcd = K.sqrt(loss_mmd + (beta * cov_dif))  # Double check sqrt --> whether we need to square the mmd
            return loss_mmcd
        elif method == 'bk':  # Means we're running mmcd on the basic version with semi kernal
            # Since the covariance matrix will just be 1's on the diag (when we have an isotropic gaussian) we can just
            # use the identity matrix
            cov_yy = self.compute_kernel(tfp.stats.covariance(latent_z_log_sigma),
                                         tfp.stats.covariance(latent_z_log_sigma))
            cov_true = tfp.stats.covariance(true_samples)
            cov_ker = self.compute_kernel(cov_true, cov_true)
            cov_dif = K.sum(K.square(cov_ker - cov_yy))

            # Since the means of a standard normal are zeros we just calculate
            # the norm of the vector of our latent dataset
            x_kernel = self.compute_kernel(true_samples, true_samples)
            y_kernel = self.compute_kernel(latent_z, latent_z)
            means_dif = K.sum(K.square(K.mean(x_kernel) - K.mean(y_kernel)))
            loss_mmcd = K.sqrt(means_dif + (beta * cov_dif))
            return loss_mmcd
        elif method == 'b':
            # Without kernel below
            # means_dif = tf.square(tf.math.reduce_mean(latent_z_mean))
            # cov_dif = K.sqrt(K.square(-1 + K.min(latent_z_log_sigma)) + K.square(1 - K.max(latent_z_log_sigma)))/2
            # loss_mmcd = (means_dif + (beta * cov_dif))
            # means_dif = tf.square(tf.math.reduce_mean(latent_z_mean))
            means_dif = tf.square(true_samples - latent_z_mean)
            cov_dif = K.sqrt(K.square(1 + K.min(latent_z_log_sigma)) + K.square(1 - K.max(latent_z_log_sigma)))/2
            loss_mmcd = (means_dif + (beta * cov_dif))
        return loss_mmcd

    def get_mmd_distance(self, latent_z):
        """
        https://github.com/Saswatm123/MMD-VAE/blob/master/MMD_VAE.ipynb
        https://learning.mpi-sws.org/mlss2016/slides/cadiz16_2.pdf
        http://abdulfatir.com/Implicit-Reparameterization/
        Parameters
        ----------
        train_latent:               latent code
        train_reconstruction:       restructured data
        train_inputs:               training data

        Returns
        -------
        new loss
        """
        batch_size = K.shape(latent_z)[0]

        latent_dim = K.int_shape(latent_z)[1]

        # Randomly sample from a normal distribution
        true_samples = K.random_normal(shape=(batch_size, latent_dim), mean=0., stddev=1.)

        # Compute loss between the training latent space and normal data
        x_kernel = self.compute_kernel(true_samples, true_samples)
        y_kernel = self.compute_kernel(latent_z, latent_z)
        xy_kernel = self.compute_kernel(true_samples, latent_z)
        loss_mmd = K.mean(x_kernel) + K.mean(y_kernel) - 2 * K.mean(xy_kernel)

        # add mmd loss and true loss
        return loss_mmd

    def get_bimodal_mmd_distance(self, latent_z):
        """ Compute MMD twice, once with each different mean. """

        batch_size = K.shape(latent_z)[0]
        dim = K.int_shape(latent_z)[1]
        zs = tf.zeros(K.shape(latent_z))#+ tf.float32
        # # Make a mask for which values should go in which tensors based on whether they are < or > 0
        # # Zero out where we don't have values matching (i.e. to make th epsilons correspond to the correct zmean
        # grp_1 = tf.math.greater(latent_z, zs)
        # grp_2 = tf.math.greater_equal(zs, latent_z)
        #
        # # Randomly sample from a normal distribution
        # #slice_1 = tf.slice(latent_z, [0, 0], [batch_size, dim])
        #
        # true_samples = K.random_normal(shape=(batch_size, dim), mean=1., stddev=1.) # Seed?
        #
        # # Compute loss between the training latent space and normal data
        # true_grp_1 = tf.boolean_mask(true_samples, grp_1)
        # latent_z_1 = tf.boolean_mask(latent_z, grp_1)
        #
        # x_kernel = self.compute_kernel(true_grp_1, true_grp_1)
        # y_kernel = self.compute_kernel(latent_z_1, latent_z_1)
        # xy_kernel = self.compute_kernel(true_grp_1, latent_z_1)
        # loss_mmd = K.mean(x_kernel) + K.mean(y_kernel) - 2 * K.mean(xy_kernel)
        #
        # # Do the same for group 2
        # # true_samples = K.random_normal(shape=(batch_size, grp_2), mean=-1., stddev=1.)
        # # slice_2 = tf.slice(latent_z, [0, grp_1], [batch_size, grp_2])
        # # Compute loss between the training latent space and normal data
        # true_grp_2 = tf.boolean_mask(true_samples, grp_2)
        # latent_z_2 = tf.boolean_mask(latent_z, grp_2)
        # x_kernel = self.compute_kernel(true_grp_2, true_grp_2)
        # y_kernel = self.compute_kernel(latent_z_2, latent_z_2)
        # xy_kernel = self.compute_kernel(true_grp_2, latent_z_2)
        # loss_mmd += K.mean(x_kernel) + K.mean(y_kernel) - 2 * K.mean(xy_kernel)
        # New loss
        grp_1 = tf.cast(tf.math.greater(latent_z, zs), tf.float32)
        grp_2 = tf.cast(tf.math.greater_equal(zs, latent_z), tf.float32)
        epsilon_1 = K.random_normal(shape=(batch_size, dim), mean=1, stddev=1)
        epsilon_2 = K.random_normal(shape=(batch_size, dim), mean=-1, stddev=1)
        # Zero out where we don't have values matching (i.e. to make th epsilons correspond to the correct zmean
        epsilon_1 = epsilon_1 * grp_1
        epsilon_2 = epsilon_2 * grp_2
        epsilon = epsilon_1 + epsilon_2  # tf.keras.layers.Concatenate(axis=1)([epsilon_1, epsilon_2])
        tue_samples = 0 + K.exp(0.5) * epsilon
        x_kernel = self.compute_kernel(tue_samples, tue_samples)
        y_kernel = self.compute_kernel(latent_z, latent_z)
        xy_kernel = self.compute_kernel(tue_samples, latent_z)
        loss_mmd = K.mean(x_kernel) + K.mean(y_kernel) - 2 * K.mean(xy_kernel)
        # add mmd loss and true loss
        return loss_mmd

    @staticmethod
    def compute_kernel(x, y):
        """
        https://stats.stackexchange.com/questions/239008/rbf-kernel-algorithm-python
        Parameters
        ----------
        x
        y

        Returns
        -------

        """
        x_size = K.shape(x)[0]
        y_size = K.shape(y)[0]
        dim = K.shape(x)[1]
        tiled_x = K.tile(K.reshape(x, [x_size, 1, dim]), [1, y_size, 1])
        tiled_y = K.tile(K.reshape(y, [1, y_size, dim]), [x_size, 1, 1])
        return K.exp(-K.mean(K.square(tiled_x - tiled_y), axis=2) / K.cast(dim, 'float32'))



###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

from scivae import VAE
from sciviso import Heatmap, Histogram, Scatterplot
from scipy import stats
from sciutil import SciUtil
import matplotlib.pyplot as plt
from sklearn.preprocessing import MinMaxScaler
from statsmodels.stats.multitest import multipletests
import pandas as pd
import os

"""
This class is build when VAE is created and allows for the plots to be generated.
"""


class Vis:

    def __init__(self, vae: VAE, sciutil: SciUtil, colours: list, config=None):
        self.vae = vae
        self.u = sciutil
        self.colours = colours if colours else ['#483873', '#1BD8A6', '#B117B7', '#AAC7E2', '#FFC107', '#016957',
                                                '#9785C0', '#D09139', '#338A03', '#FF69A1', '#5930B1', '#FFE884',
                                                '#35B567', '#1E88E5', '#ACAD60', '#A2FFB4', '#B618F5', '#854A9C']
        self.df = None
        self.base_config = {
            'font': {'family': 'normal', 'size': 6},
            'label_font_size': 7,
            'title_font_size': 10,
            'style': "ticks",
            'figsize': (3, 3),
            'alpha': 0.6
        }
        self.base_config = self.__merge_config(self.base_config, config)

    def __merge_config(self, default_config, user_config):
        # Add in default params from our base
        for c in self.base_config:
            if not default_config.get(c):
                default_config[c] = self.base_config[c]
        if user_config:
            for c in default_config:
                # Update any of the config if the user has supplied it
                user_config[c] = user_config.get(c) if user_config.get(c) is not None else default_config.get(c)
            return user_config
        return default_config

    def plot_input_distribution(self, df: pd.DataFrame, row_id="", columns=None, output_dir="",
                                fig_type='svg', title="Hist", show_plt=False, save_fig=True, user_config=None):
        columns = columns if columns is not None else [c for c in df.columns if c != row_id]
        config = {
            'figsize': (4, 4),
            'nbins': 20,
            'c': "grey",
            'max_y': None,
            'title': title
        }
        config = self.__merge_config(config, user_config)
        for column in columns:
            set_style()
            try:
                hist = Histogram(df, column, title=column.replace('_', ' '), config=config)
                hist.plot()
                if save_fig:
                    plt.savefig(os.path.join(output_dir, f'{title}_{column}.{fig_type}'))
                if show_plt:
                    plt.show()
            except:
                self.u.warn_p(["plot_input_distribution: Unable to plot histogram for column:", column])

    def plot_feature_correlation(self, df: pd.DataFrame, row_id="", columns=None, output_dir="",
                                 fig_type='svg', print_vals=True, title="RCM heatmap of feature correlation",
                                 show_plt=False, save_fig=True, user_config=None):
        heatmap_df = pd.DataFrame()
        config = {
            'figsize': (4, 4),
            'cluster_cols': True,
            'cluster_rows': True,
            'vmin': -1,
            'vmax': 1,
            'cmap': 'RdBu_r',
            'title': title
        }
        config = self.__merge_config(config, user_config)

        lbls = []
        padj_cols = []
        columns = columns if columns is not None else [c for c in df.columns if c != row_id]
        for feature_1 in columns:
            vals = []
            pvals = []
            for feature_2 in columns:
                rho, p = stats.spearmanr(df[feature_1].values, df[feature_2].values)
                if print_vals:
                    print(feature_1, feature_2, rho, p)
                vals.append(rho)
                pvals.append(p)
            lbls.append(f'{feature_1}')
            heatmap_df[f'{feature_1}'] = vals
            reg, padj, a, b = multipletests(pvals, alpha=0.1, method='fdr_bh', returnsorted=False)
            heatmap_df[f'{feature_1} padj'] = [float(f'{p:.3f}') for p in padj]
            # heatmap_df[f'Node {i + 1} padj'] = [float(f'rho:{vals[ip]:.3f} p:{p:.3f}') for ip, p in enumerate(padj)]
            padj_cols.append(f'{feature_1} padj')
        heatmap_df['labels'] = lbls
        # annot_values = heatmap_df[padj_cols].values
        heatmap = Heatmap(heatmap_df, lbls, 'labels', config=config)
        g = heatmap.plot()
        if save_fig:
            plt.savefig(os.path.join(output_dir, f'Heatmap_{title.replace(" ", "")}.{fig_type}'))
        if show_plt:
            plt.show()
        return heatmap_df

    def plot_top_values_by_rank(self, df: pd.DataFrame, rank_colums: list, vis_columns: list, id_col: str,
                                num_values=10, fig_type='svg', output_dir="", title="Ranked top values",
                                show_plt=False, save_fig=True, user_config=None):
        """
        Basically, for each of the rank columns we want to plot the values (from value filter) and the
        top and bottom of the ranks. We plot the vis columns for each of these. Note the vis columns
        should all have the same range.
        """
        config = {
            'figsize': (4, 4),
            'cluster_cols': True,
            'cluster_rows': True,
            'vmin': -1,
            'vmax': 1,
            'cmap': 'RdBu_r',
            'title': title,
        }
        config = self.__merge_config(config, user_config)
        rank_colours = []
        order_colours = []
        value_cols = []
        heatmap_df = pd.DataFrame()
        for v_c in vis_columns:
            values = []
            gs = []

            for r_i, r_c in enumerate(rank_colums):
                # First add the bottom ones
                rank_values = df[r_c].values
                desc_sorted = (-1 * rank_values).argsort()  # Sort the values by descending order
                ids_sorted = df[id_col].values[desc_sorted]
                data_sorted = df[v_c].values[desc_sorted]
                i = 0
                for g in ids_sorted:
                    gs.append(g)
                    rank_colours.append(self.colours[r_i])
                    order_colours.append("red")
                    i += 1
                    if i >= num_values:
                        break
                values += list(data_sorted[:num_values])
                # Do the same with the top values
                rank_values = df[r_c].values
                desc_sorted = (rank_values).argsort()  # Sort the values by descending order
                ids_sorted = df[id_col].values[desc_sorted]
                data_sorted = df[v_c].values[desc_sorted]
                i = 0
                for g in ids_sorted:
                    gs.append(g)
                    rank_colours.append(self.colours[r_i])
                    order_colours.append("blue")
                    i += 1
                    if i >= num_values:
                        break
                values += list(data_sorted[:num_values])

            heatmap_df[v_c] = values
            value_cols.append(v_c)
        heatmap_df[id_col] = gs  # This should be the same irrespective since the
        heatmap = Heatmap(heatmap_df, value_cols, id_col, row_colours=[rank_colours, order_colours], title=title,
                          config=config)
        heatmap.plot()
        if save_fig:
            plt.savefig(os.path.join(output_dir, f'Heatmap_{title.replace(" ", "_")}.{fig_type}'))
        if show_plt:
            plt.show()
        plt.show()

    def plot_node_correlation(self, output_dir="", fig_type='svg', print_vals=True, title="RCM heatmap of node cor.",
                              show_plt=False, save_fig=True, user_config=None):
        heatmap_df = pd.DataFrame()
        config = {
            'figsize': (4, 4),
            'cluster_cols': True,
            'cluster_rows': True,
            'vmin': -1,
            'vmax': 1,
            'cmap': 'RdBu_r',
            'title': title,
        }
        config = self.__merge_config(config, user_config)
        lbls = []
        if not self.vae:
            self.u.err_p(["plot_node_correlation: You haven't run the VAE yet! "
                          "Please run compute_vae before attempting to plot."])
            return
        data = self.vae.get_encoded_data()

        padj_cols = []
        num_nodes = len(data[0])
        for i in range(0, num_nodes):
            vals = []
            pvals = []
            for j in range(0, num_nodes):
                rho, p = stats.spearmanr(data[:, i], data[:, j])
                if print_vals:
                    print(i + 1, j + 1, rho, p)
                vals.append(rho)
                pvals.append(p)
            lbls.append(f'Node {i + 1}')
            heatmap_df[f'Node {i + 1}'] = vals
            reg, padj, a, b = multipletests(pvals, alpha=0.1, method='fdr_bh', returnsorted=False)
            heatmap_df[f'Node {i + 1} padj'] = [float(f'{p:.3f}') for p in padj]
            padj_cols.append(f'Node {i + 1} padj')
        heatmap_df['labels'] = lbls

        # annot_values = heatmap_df[padj_cols].values
        set_style()
        heatmap = Heatmap(heatmap_df, lbls, 'labels', config=config)
        heatmap.plot()
        if save_fig:
            plt.savefig(os.path.join(output_dir, f'Heatmap_{title.replace(" ", "_")}.{fig_type}'))
        if show_plt:
            plt.show()
        return heatmap_df

    def plot_node_feature_correlation(self, df: pd.DataFrame, row_id: str, vae_data=None, columns=None, output_dir="",
                                      fig_type='svg', print_vals=True, encoding_type="z",
                                      title="RCM heatmap of node feature corr", show_plt=False, save_fig=True,
                                      user_config=None):
        set_style()
        config = {
            'figsize': (4, 4),
            'cluster_cols': True,
            'cluster_rows': True,
            'vmin': -1,
            'vmax': 1,
            'cmap': 'RdBu_r',
            'title': title,
        }
        config = self.__merge_config(config, user_config)
        heatmap_df = pd.DataFrame()
        lbls, cols = [], []
        columns = columns if columns is not None else [c for c in df.columns if c != row_id]
        if not self.vae:
            self.u.err_p(["plot_scatters: You haven't run the VAE yet! "
                          "Please run compute_vae before attempting to plot."])
        if vae_data is None:
            scaler = MinMaxScaler(copy=True)
            data_cols = [c for c in df.columns if c != row_id]
            dataset = df[data_cols].values
            scaled_vals = scaler.fit_transform(dataset)
            data = self.vae.encode_new_data(scaled_vals, encoding_type=encoding_type)
        else:
            data = vae_data
        num_nodes = len(data[0])
        first = True
        padj_cols = []
        for i in range(0, num_nodes):
            vals = []
            pvals = []
            for column in columns:
                if first:
                    lbls.append(f'{column}')
                rho, p = stats.spearmanr(data[:, i], df[column].values)
                if print_vals:
                    print(i + 1, column, rho, p)
                vals.append(rho)
                pvals.append(p)
            cols.append(f'Node {i + 1}')
            heatmap_df[f'Node {i + 1}'] = vals
            reg, padj, a, b = multipletests(pvals, alpha=0.1, method='fdr_bh', returnsorted=False)
            heatmap_df[f'Node {i + 1} padj'] =  [float(f'{p:.3f}') for p in padj]
            padj_cols.append(f'Node {i + 1} padj')
            first = False
        heatmap_df['labels'] = lbls
        # annot_values = heatmap_df[padj_cols].values
        set_style()
        heatmap = Heatmap(heatmap_df, cols, 'labels', config=config)
        heatmap.plot()
        if save_fig:
            plt.savefig(os.path.join(output_dir, f'Heatmap_{title.replace(" ", "_")}.{fig_type}'))
        if show_plt:
            plt.show()
        return heatmap_df

    def plot_node_hists(self, title="Hist", output_dir="", fig_type='svg', method="z", show_plt=False, save_fig=True,
                        user_config=None):
        if not self.vae:
            self.u.err_p(["plot_node_correlation: You haven't run the VAE yet! "
                          "Please run compute_vae before attempting to plot."])
            return
        config = {
            'figsize': (4, 4),
            'nbins': 20,
            'c': "grey",
            'max_y': None,
            'title': title
        }
        config = self.__merge_config(config, user_config)
        set_style()
        data = self.vae.get_encoded_data(method)
        num_nodes = len(data[0])
        for i in range(0, num_nodes):
            set_style()
            tmp_df = pd.DataFrame()
            tmp_df[f'Node {i + 1}'] = data[:, i]
            config['title'] = f'{title} Node {i + 1}'
            hist = Histogram(tmp_df, f'Node {i + 1}', config=config)
            hist.plot()

            if save_fig:
                plt.savefig(os.path.join(output_dir, f'{title}_Node-{i + 1}.{fig_type}'))
            if show_plt:
                plt.show()
            plt.clf()

    @staticmethod
    def get_row_idxs(df, row_id, row_labels, row_values):
        row_idxs = {}
        for m in row_labels:
            row_idxs[m] = []

        g_i = 0
        for g in df[row_id].values:
            i = 0
            for m in row_labels:
                if g in row_values[i]:
                    row_idxs[m].append(g_i)

                i += 1
            g_i += 1
        row_idxs_lst = []
        for g in row_labels:
            row_idxs_lst.append(row_idxs.get(g))
        return row_idxs_lst

    def plot_feature_scatters(self, df: pd.DataFrame, row_id: str, vae_data=None, columns=None, output_dir="",
                              fig_type='svg', title="latent space", show_plt=False, save_fig=True, encoding_type="z",
                              angle_plot=90, user_config=None):
        set_style()
        config = {
            'figsize': (4, 4),
            'angle_plot': angle_plot,
            'c': "grey",
            'max_y': None,
            'title': title
        }
        config = self.__merge_config(config, user_config)
        columns = columns if columns is not None else [c for c in df.columns if c != row_id]
        if not self.vae:
            self.u.err_p(["plot_scatters: You haven't run the VAE yet! "
                          "Please run compute_vae before attempting to plot."])
        if vae_data is None:
            scaler = MinMaxScaler(copy=True)
            data_cols = [c for c in df.columns if c != row_id]
            dataset = df[data_cols].values
            scaled_vals = scaler.fit_transform(dataset)
            data = self.vae.encode_new_data(scaled_vals, encoding_type=encoding_type)
        else:
            data = vae_data
        num_nodes = len(data[0])
        if num_nodes > 3:
            self.u.warn_p(["plot_scatters", "You have more than three latent nodes: ", num_nodes,
                           "\n We will only be plotting the first three"])
        if num_nodes < 2:
            self.u.warn_p(["plot_scatters", "You have less than two nodes: ", num_nodes,
                           "\nWe won't plot this ... returning"])
            return
        for column in columns:
            vis_df = pd.DataFrame()
            vis_df[row_id] = df[row_id].values
            vis_df[f'Dim. 1'] = data[:, 0]
            vis_df[f'Dim. 2'] = data[:, 1]
            if num_nodes == 2:
                scatter = Scatterplot(vis_df, f'Dim. 1', f'Dim. 2', f'{column} {title}', f'Dim. 1', f'Dim. 2',
                                      colour=pd.to_numeric(df[column].values), config=config)
            else:
                vis_df[f'Dim. 3'] = data[:, 2]
                scatter = Scatterplot(vis_df, f'Dim. 1', f'Dim. 2', f'{column} {title}', f'Dim. 1', f'Dim. 2',
                                      z=f'Dim. 3', colour=pd.to_numeric(df[column].values), config=config)
            ax = scatter.plot()
            if save_fig:
                plt.savefig(os.path.join(output_dir, f'Scatter_{column}_{title.replace(" ", "_")}.{fig_type}'))
            if angle_plot:
                # Plot different angles
                for ii in range(0, 360, angle_plot):
                    ax.view_init(elev=10., azim=ii)
                    plt.savefig(os.path.join(output_dir, f'Scatter_{column}_{title.replace(" ", "_")}_{ii}.{fig_type}'))
            if show_plt:
                plt.show()

    def plot_values_on_scatters(self, df: pd.DataFrame, row_id: str, row_labels: list, row_values: list, vae_data=None,
                                output_dir="", fig_type='svg', title="RCM genes on latent space", show_plt=False,
                                color_map=None, save_fig=True, angle_plot=None, encoding_type="z", user_config=None):
        set_style()
        config = {
            'figsize': (4, 4),
            'angle_plot': angle_plot,
            'c': "grey",
            'max_y': None,
            'title': title,
            'vmin': None,
            'vmax': None,
            'plt_bg': False
        }
        config = self.__merge_config(config, user_config)
        marker_idxs = self.get_row_idxs(df, row_id, row_labels, row_values)

        if vae_data is None:
            if not self.vae:
                self.u.err_p(["plot_scatters: You haven't run the VAE yet! "
                              "Please run compute_vae before attempting to plot."])
            scaler = MinMaxScaler(copy=True)
            data_cols = [c for c in df.columns if c != row_id]
            dataset = df[data_cols].values
            scaled_vals = scaler.fit_transform(dataset)
            data = self.vae.encode_new_data(scaled_vals, encoding_type=encoding_type)
        else:
            data = vae_data
        num_nodes = len(data[0])
        if num_nodes > 3:
            self.u.warn_p(["plot_scatters", "You have more than three latent nodes: ", num_nodes,
                           "\n We will only be plotting the first three"])
        if num_nodes < 2:
            self.u.warn_p(["plot_scatters", "You have less than two nodes: ", num_nodes,
                           "\nWe won't plot this ... returning"])
            return
        i = 0
        if color_map is None:
            color_map = []
            for c in row_labels:
                color_map.append(self.colours[i])
                i += 1
                if i > len(self.colours):
                    i = 0

        vis_df = pd.DataFrame()
        vis_df[row_id] = df[row_id].values
        vis_df[f'Dim. 1'] = data[:, 0]
        vis_df[f'Dim. 2'] = data[:, 1]
        if num_nodes == 2:
            scatter = Scatterplot(vis_df, f'Dim. 1', f'Dim. 2', f'{title}', f'Dim. 1', f'Dim. 2', config=config)
            ax = scatter.plot_groups_2D(row_labels, marker_idxs, color_map, alpha_bg=0.1)
        else:
            vis_df[f'Dim. 3'] = data[:, 2]
            scatter = Scatterplot(vis_df, f'Dim. 1', f'Dim. 2', f'{title}', f'Dim. 1', f'Dim. 2', z=f'Dim. 3',
                                  config=config)
            ax = scatter.plot_groups_3D(row_labels, marker_idxs, color_map, alpha_bg=0.1)
        if save_fig:
            plt.savefig(os.path.join(output_dir, f'Scatter-genes_{title.replace(" ", "_")}.{fig_type}'))
        if angle_plot:
            # Plot different angles
            for ii in range(0, 360, angle_plot):
                ax.view_init(elev=10., azim=ii)
                plt.savefig(os.path.join(output_dir, f'Scatter-genes_{title.replace(" ", "_")}_{ii}.{fig_type}'))
        if show_plt:
            plt.show()
        set_style()
        return ax


def set_style():
    # plot
    plt.clf()
